import logging
import typing

from datetime import datetime

from .context import get_http_request, get_labels, get_span_id, utcnow
from .types import LogEntry, LogLevel, SourceLocation


LabelOrGetter = typing.Union[str, typing.Callable[[logging.LogRecord], str]]


class GKELoggingFormatter(logging.Formatter):
    """
    Standard Python logging.Formatter for implementing Google Cloud's "structured logging" in GKE
    containers.

    This implementation uses optional contextvars to set some additional fields to further color
    each log entry's context:

        - spanId via set_span_id() and get_span_id()
        - httpRequest via set_http_request() and get_http_request()

    This implementation also provides for a set of default metadata labels to be defined for
    app-wide log annotations, which should be used for labeling log messages with information such
    as application identifiers, version numbers, environments, etc. These default labels can be
    expressed either as static strings, or functions that can compute a value given a LogRecord

    See https://cloud.google.com/logging/docs/structured-logging
    """

    def __init__(self, default_labels: typing.Mapping[str, LabelOrGetter] = {}):
        super().__init__()
        self._default_labels = default_labels

    def _get_labels(self, record: logging.LogRecord) -> typing.Mapping[str, str]:
        """
        Computes the labels for this log entry by merging labels in the below order:

          - "Default labels" configured on the formatter instance
          - "Contextual labels" configured in the current task context
          - LogRecord "extras" set at point of the logger use
        """
        default_labels = {
            k: v(record) if callable(v) else v for k, v in self._default_labels.items()
        }
        return {
            **default_labels,
            **get_labels(),
            **record.__dict__.get("labels", {}),
        }

    def format(self, record: logging.LogRecord) -> str:
        timestamp: datetime = record.__dict__.get("timestamp", utcnow())
        log_entry = LogEntry(
            time=timestamp.isoformat(),
            severity=LogLevel[record.levelname],
            message=super().format(record),
            http_request=get_http_request(),
            source_location=SourceLocation(
                file=record.pathname,
                line=record.lineno,
                function=record.funcName,
            ),
            span_id=get_span_id(),
            labels=self._get_labels(record),
        )
        # Output the JSON by pydantic field "alias" so that we can use non-field-friendly output
        # keys
        return log_entry.json(by_alias=True, exclude_none=True)



import logging
import sys
import time
import typing

from datetime import datetime
from starlette.datastructures import Headers
from starlette.types import ASGIApp, Message, Receive, Send, Scope
from starlette.requests import Request

from .context import get_user_id, set_http_request, utcnow
from .pylogging import GKELoggingFormatter
from .types import HttpRequest


def build_http_request_from_scope(scope: Scope) -> HttpRequest:
    """
    Converts an ASGI scope into a partially-complete HttpRequest object suitable for GKE structured
    JSON logging
    """

    req = Request(scope)
    server_ip, _ = req.scope["server"]
    protocol = f"HTTP/{req.scope['http_version']}"
    return HttpRequest(
        protocol=protocol,
        method=req.method,
        url=str(req.url),
        # TODO: Confirm if this is accurate even for body-less requests
        request_size=req.headers.get("Content-Length"),
        user_agent=req.headers["User-Agent"],
        remote_ip=req.client.host,
        server_ip=server_ip,
        referer=req.headers.get("Referer"),
    )


class GKELoggingMiddleware:
    """
    ASGI middleware for logging access logs via Google Cloud's "structured logging" in GKE
    containers.

    Under the hood, this works by leveraging the GKELoggingFormatter and setting the httpRequest via
    contextvar scoped to each incoming request.
    """

    """
    This should reflect the "combined" common log format, commonly found in major web servers like
    Apache httpd and NGINX (https://httpd.apache.org/docs/trunk/logs.html#combined).

    The format generally contains these atoms:
        - Remote address (client IP)
        - RFC-1413 identity
        - User ID
        - Request timestamp
        - HTTP request line (i.e. "<METHOD> <REQUEST_URL> <PROTOCOL>")
        - HTTP response status code
        - HTTP response length in bytes
        - HTTP referer ("combined")
        - User agent ("combined")

    Note that a dash is used to indicate any missing data
    """
    common_log_format = " ".join(
        (
            "{remote_addr}",
            "{ident}",
            "{user_id}",
            "[{timestamp}]",
            "{request_line}",
            "{status_code}",
            "{response_length}",
            '"{referer}"',
            '"{user_agent}"',
        )
    )

    def __init__(
        self,
        app: ASGIApp,
        access_log: typing.Union[logging.Logger, str] = "asgi.access",
        access_log_message_format: str = common_log_format,
        **formatter_args,
    ):
        self._app = app
        if isinstance(access_log, str):
            self._logger = logging.Logger(access_log, level=logging.INFO)
        else:
            self._logger = access_log
        h = logging.StreamHandler(sys.stdout)
        h.setFormatter(GKELoggingFormatter(**formatter_args))
        self._logger.addHandler(h)
        self._access_log_message_format = access_log_message_format

    def log_request(self, request_time: datetime, http_request: HttpRequest):
        """
        Logs out the current request using the configured access_log_message_format
        """

        timestamp = request_time.strftime("%d/%b/%Y:%H:%M:%S %z")
        url_with_query: str = http_request.url.path or "/"
        if http_request.url.query:
            url_with_query += f"?{http_request.url.query}"
        data = dict(
            remote_addr=http_request.remote_ip,
            ident=None,
            user_id=get_user_id(),
            timestamp=timestamp,
            request_line=f"{http_request.method} {url_with_query} {http_request.protocol}",
            status_code=http_request.status,
            response_length=http_request.response_size,
            referer=http_request.referer,
            user_agent=http_request.user_agent,
        )
        data = {k: v or "-" for k, v in data.items()}
        log_line = self._access_log_message_format.format(**data)

        # Log levels should reflect status code
        if http_request.status is None or http_request.status < 400:
            self._logger.info(log_line, extra=dict(timestamp=request_time))
        elif http_request.status < 500:
            self._logger.warning(log_line, extra=dict(timestamp=request_time))
        else:
            self._logger.error(log_line, extra=dict(timestamp=request_time))

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            return await self._app(scope, receive, send)  # pragma: no cover

        request_time = utcnow()
        http_request = build_http_request_from_scope(scope)
        set_http_request(http_request)

        async def send_wrapper(message: Message) -> None:
            """
            Wraps outgoing response stream messages in order to capture response line and header
            data that is important for logging in the httpRequest structured field
            """
            if message["type"] == "http.response.start":  # Capture response data
                http_request.status = message["status"]
                response_headers = Headers(raw=message.get("headers", {}))
                # TODO: Confirm this behavior for streaming responses
                http_request.response_size = response_headers.get("content-length")
            await send(message)

        start_time = time.time()
        try:
            # Continue along to the next set of middleware or endpoint, but with the send_wrapper
            # instead
            await self._app(scope, receive, send_wrapper)
        except Exception:
            # Set response status explicitly in the case of an uncaught exception
            # TODO: Confirm this is sane default behavior
            http_request.status = 500
            raise
        finally:
            # Finally record the total response latency and log the request to the access log
            end_time = time.time()
            http_request.latency = f"{end_time - start_time:.5f}s"
            self.log_request(request_time, http_request)



import typing

from datetime import datetime
from enum import Enum
from pydantic import AnyHttpUrl, BaseModel, Field


class LogLevel(str, Enum):
    """
    Typed str/Enum to match Google Cloud Logging's LogSeverity

    See https://cloud.google.com/logging/docs/reference/v2/rest/v2/LogEntry#LogSeverity
    """

    DEFAULT = "DEFAULT"
    DEBUG = "DEBUG"
    INFO = "INFO"
    NOTICE = "NOTICE"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"
    ALERT = "ALERT"
    EMERGENCY = "EMERGENCY"


class HttpMethod(str):
    """
    Typed str to match valid HTTP methods
    """

    ALLOWED_METHODS = (
        "get",
        "post",
        "patch",
        "put",
        "delete",
        "options",
        "head",
        "connect",
    )

    @classmethod
    def __get_validators__(clss):
        yield clss.validate

    @classmethod
    def validate(clss, v):
        if v.lower() not in HttpMethod.ALLOWED_METHODS:
            raise ValueError(f"Invalid HTTP method: {v}")
        return clss(v)


class HttpRequest(BaseModel):
    """
    Typed model to match Google Cloud Logging's HttpRequest

    See https://cloud.google.com/logging/docs/reference/v2/rest/v2/LogEntry#httprequest
    """

    protocol: str
    method: HttpMethod = Field(alias="requestMethod")
    url: AnyHttpUrl = Field(alias="requestUrl")
    request_size: typing.Optional[str] = Field(None, alias="requestSize")
    user_agent: str = Field(alias="userAgent")
    remote_ip: str = Field(alias="remoteIp")
    server_ip: str = Field(alias="serverIp")
    referer: typing.Optional[str] = None
    status: typing.Optional[int] = None
    response_size: typing.Optional[str] = Field(None, alias="responseSize")
    latency: typing.Optional[str] = None

    class Config:
        allow_population_by_field_name = True


class SourceLocation(BaseModel):
    """
    Typed model to match Google Cloud Logging's LogEntrySourceLocation

    See https://cloud.google.com/logging/docs/reference/v2/rest/v2/LogEntry#LogEntrySourceLocation
    """

    file: str
    line: str
    function: str

    class Config:
        allow_population_by_field_name = True


class LogEntry(BaseModel):
    """
    Typed model to match Google Cloud Logging's LogEntry via "special payload fields"

    See https://cloud.google.com/logging/docs/structured-logging#special-payload-fields
    See https://cloud.google.com/logging/docs/reference/v2/rest/v2/LogEntry
    """

    time: datetime
    severity: LogLevel
    message: str
    http_request: typing.Optional[HttpRequest] = Field(alias="httpRequest")
    span_id: typing.Optional[str] = Field(None, alias="logging.googleapis.com/spanId")
    source_location: typing.Optional[SourceLocation] = Field(
        None, alias="logging.googleapis.com/sourceLocation"
    )
    labels: typing.Mapping[str, str] = Field({}, alias="logging.googleapis.com/labels")

    class Config:
        allow_population_by_field_name = True



import typing

from contextvars import ContextVar
from datetime import datetime, timezone

from .types import HttpRequest


LABELS: ContextVar[typing.Dict[str, str]] = ContextVar("labels")
USER_ID: ContextVar[str] = ContextVar("user_id")
SPAN_ID: ContextVar[str] = ContextVar("span_id")
REQUEST: ContextVar["HttpRequest"] = ContextVar("request")


def utcnow() -> datetime:
    """
    Helper for getting the current time in UTC
    """
    return datetime.now(tz=timezone.utc)


def set_user_id(user_id: str) -> None:
    """
    Sets the user ID for the current logging context
    """
    USER_ID.set(user_id)


def get_user_id() -> typing.Optional[str]:
    """
    Retrieves the current user ID, if it exists
    """
    return USER_ID.get(None)


def add_label(label: str, value: str) -> None:
    """
    Adds an arbitrary label to the current logging context
    """
    labels = LABELS.get({})
    labels[label] = value
    LABELS.set(labels)


def set_labels(**labels) -> None:
    """
    Sets multiple labels for the current logging context
    """
    LABELS.set(labels)


def get_labels() -> typing.Mapping[str, str]:
    """
    Retrieves the current labels, or an empty dict if none have been set
    """
    return LABELS.get({})


def set_span_id(span_id: str) -> None:
    """
    Sets the span ID for the current logging context
    """
    SPAN_ID.set(span_id)


def get_span_id() -> typing.Optional[str]:
    """
    Retrieves the current span ID, if it exists
    """
    return SPAN_ID.get(None)


def set_http_request(http_request: HttpRequest) -> None:
    """
    Sets the HttpRequest for the current logging context
    """
    REQUEST.set(http_request)


def get_http_request() -> typing.Optional[HttpRequest]:
    """
    Retrieves the current HttpRequest, if it exists
    """
    return REQUEST.get(None)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gke_logging']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.0,<2.0.0']

extras_require = \
{'asgi': ['starlette>=0.18.0,<0.19.0']}

setup_kwargs = {
    'name': 'gke-logging',
    'version': '0.0.3',
    'description': 'Utilities for interacting with logging facilities in GKE workloads',
    'long_description': '# gke-logging\n\n[![PyPI version](https://badge.fury.io/py/gke-logging.svg)](https://badge.fury.io/py/gke-logging)\n\nUtilities for interacting with logging facilities in GKE workloads\n\n## Installation\n\n### Requirements\n\n- Python 3.7+\n- [Poetry](https://python-poetry.org/) (for development only)\n\n### Install from PyPI (recommended)\n\n```\npip install gke-logging\n```\n\n### Installing from Github\n\n```\npip install git+https://github.com/StationA/gke-logging.git#egg=gke-logging\n```\n\n### Installing from source\n\n```\ngit clone https://github.com/StationA/gke-logging.git\ncd gke-logging\npoetry install\n```\n\n## Usage\n\n### `gke_logging.GKELoggingFormatter`\n\nOne of the core components is the `GKELoggingFormatter`, which is an implementation of the built-in\n`logging.Formatter` protocol that translates a `logging.LogRecord` into a JSON format that GKE\'s\nlogging infrastructure can understand. At a minimum, this enables any software running on GKE to\nintegrate structured logging simply by applying this formatter for your loggers, e.g.:\n\n```python\nimport logging\n\nfrom gke_logging import GKELoggingFormatter\n\n\nLOGGER = logging.getLogger(__name__)\nh = logging.StreamHandler()\nh.setFormatter(GKELoggingFormatter())\nLOGGER.addHandler(h)\nLOGGER.setLevel(logging.INFO)\n\n\n# ...\n\nLOGGER.info("Look at me! I can haz GKE structured logging!")\n# Prints out: {"time": "2022-01-13T23:22:26.336686+00:00", "severity": "INFO", "message": "Look at me! I can haz GKE structured logging!", "logging.googleapis.com/sourceLocation": {"file": "test_log.py", "line": "14", "function": "<module>"}, "logging.googleapis.com/labels": {}}\n```\n\nFurthermore, this formatter allows you to set app-level metadata to be sent along with each log\nmessage, which is useful in order to better organize collected log data:\n\n```python\n# ...\nh.setFormatter(\n    GKELoggingFormatter(default_labels=dict(app_id="my-cool-app", version="0.1.0"))\n)\n# ...\n```\n\nAlso the formatter also allows you to add HTTP metadata to any logs that occur during the course of\na request. This enhances logs that are emitted during request-handling logic in APIs with additional\ndata. This functionality is primarily utilized in the included `GKELoggingMiddleware` in order to\nprovide basic access logs.\n\n### `gke_logging.asgi.GKELoggingMiddleware`\n\n`gke_logging.asgi.GKELoggingMiddleware` is an ASGI middleware that emits basic access logs in\n"common log format", with a default behavior that integrates with the `GKELoggingFormatter` to write\nthe access logs in a format that GKE\'s logging infrastructure better understands. By implementing\nper the ASGI spec, this means it can work with any ASGI-compatible server, including FastAPI,\nstarlette, and ASGI implementations:\n\n```python\nfrom fastapi import FastAPI\nfrom gke_logging.asgi import GKELoggingMiddleware\n\napp = FastAPI()\napp.add_middleware(GKELoggingMiddleware)\n\n@app.get("/")\ndef get_it() -> str:\n    return "OK"\n```\n\nAdditionally, because this middleware integrates with `gke_logging.context` bindings, it enables any\nlogger used during the course of handling a request to emit logs that also contain request-time\ndata, e.g. request URL, user-agent, response latency, etc.\n\n```python\nimport logging\n\nfrom fastapi import FastAPI\nfrom gke_logging import GKELoggingFormatter\nfrom gke_logging.asgi import GKELoggingMiddleware\n\napp = FastAPI()\napp.add_middleware(GKELoggingMiddleware)\n\n\nroot_logger = logging.getLogger()\nh = logging.StreamHandler()\nh.setFormatter(GKELoggingFormatter())\nroot_logger.setLevel(logging.INFO)\nroot_logger.addHandler(h)\n\n\n@app.get("/")\ndef get_it() -> str:\n    # Any log records created during request-handling will be enriched with other HTTP request data\n    root_logger.info("TEST")\n    return "OK"\n```\n\n### `gke_logging.context`\n\nIn order to control additional metadata labels for log records that correspond to one logical\noperation, e.g. an HTTP request, a batch job operation, etc., you should use the helper functions\nexported in `gke_logging.context`:\n\n```python\nimport logging\n\nfrom contextvars import copy_context\n\nfrom gke_logging import GKELoggingFormatter\nfrom gke_logging.context import set_labels\n\n\nLOGGER = logging.getLogger(__name__)\nh = logging.StreamHandler()\nh.setFormatter(GKELoggingFormatter())\nLOGGER.addHandler(h)\nLOGGER.setLevel(logging.INFO)\n\n# ...\n\n\ndef run_job(job_id: str):\n    set_labels(job_id=job_id)\n    LOGGER.info("TEST")\n\n\nctx = copy_context()\nfor i in range(10):\n    ctx.run(run_job, f"{i + 1}")\n```\n\nBecause `ContextVar`s bind natively to Python\'s `asyncio`, you can use these same helper\nfunctions within asynchronous tasks in a similar fashion.\n\n### Additional examples\n\nAdditional usage examples can be found in [examples/](examples/)\n\n## Contributing\n\nWhen contributing to this repository, please follow the steps below:\n\n1. Fork the repository\n1. Submit your patch in one commit, or a series of well-defined commits\n1. Submit your pull request and make sure you reference the issue you are addressing\n',
    'author': 'Station A',
    'author_email': 'oss@stationa.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
